#include <ace/OS.h>
#include <ace/VRT_Wrapper.h>
#include <ace/Virtual_Real_Time.h>

#if defined (ACE_HAS_VIRTUAL_TIME)

#if !defined (__ACE_INLINE__)
#include "ace/VRT_wrapper.inl"
#endif /* __ACE_INLINE__ */

// time functions return Virtual Time
ACE_Time_Value 
ACE_VRT_Wrapper::gettimeofday (void)
{ 
  if (!ACE_Virtual_Real_Time::instance()->in_test_group())
    return ACE_OS::gettimeofday_i ();

  return ACE_Virtual_Real_Time::instance()->current_virtual_time();
}

int
ACE_VRT_Wrapper::cond_timedwait (ACE_cond_t *cv,
                                 ACE_mutex_t *m,
                                 ACE_Time_Value *timeout)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if (timeout == NULL || !vrt->in_test_group())
    return ACE_OS::cond_timedwait_i (cv, m, timeout);

  int ret_val;
  
  ACE_Time_Value expected_end (*timeout);
  ACE_Time_Value time_to_wait (expected_end - vrt->current_virtual_time());
  time_to_wait = vrt->scaled_time_value (time_to_wait);
  while (1)
    {
      ACE_Time_Value real_time_to_wait = time_to_wait +
	ACE_VRT_Wrapper::getrealtimeofday ();

      ret_val = ACE_OS::cond_timedwait_i (cv, m, &real_time_to_wait);
      
      if (ret_val == 0 ||       // call succeeds
          errno != ETIME ||     // error other than timeout
          vrt->wait_longer(expected_end, &time_to_wait) == 0) 
	break;
    }
  return ret_val; 
}

int
ACE_VRT_Wrapper::sleep (const ACE_Time_Value &tv)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if (!vrt->in_test_group())
    return ACE_OS::sleep_i (tv);
  
  ACE_Time_Value time_to_wait (vrt->scaled_time_value(tv));
  ACE_Time_Value end_time = vrt->current_virtual_time() + tv;
  while (1)
    {
     // We are not using a sleep implementation here since ordinary
      // sleep does not have sufficient resolution.
      // Here the only thing that can happen is a time-out.
      int ret_val = ACE_OS::select_i (0, NULL, NULL, NULL, &time_to_wait); 

      if (ret_val == -1 || vrt->wait_longer(end_time,&time_to_wait) == 0)
	return ret_val;
    }
  return 0;
}

int
ACE_VRT_Wrapper::sleep (u_int seconds)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if (!vrt->in_test_group())
    return ACE_OS::sleep_i (seconds);

  ACE_Time_Value time_to_wait (seconds);
  ACE_Time_Value end_time = vrt->current_virtual_time() + time_to_wait;
  time_to_wait = vrt->scaled_time_value (time_to_wait);
  while (1)
    {
     // We are not using a sleep implementation here since ordinary
      // sleep does not have sufficient resolution.
      // Here the only thing that can happen is a time-out.
      int ret_val = ACE_OS::select_i (0, NULL, NULL, NULL, &time_to_wait); 

      if (ret_val == -1 || vrt->wait_longer(end_time,&time_to_wait) == 0)
	return ret_val;
    }
  return 0;
}

int
ACE_VRT_Wrapper::nanosleep (const struct timespec *requested,
                            struct timespec *remaining)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if (!vrt->in_test_group())
    return ACE_OS::nanosleep_i (requested, remaining);

  ACE_Time_Value time_to_wait (requested->tv_sec, requested->tv_nsec / 1000);
  ACE_Time_Value end_time = vrt->current_virtual_time() + time_to_wait;
  time_to_wait = vrt->scaled_time_value (time_to_wait);
  
  timespec wait_time;
  while (1)
    {
      wait_time = timespec_t (time_to_wait);      
      int ret_val = ACE_OS::nanosleep_i (&wait_time, remaining);
      
      //call fails
      if (ret_val == -1 || vrt->wait_longer(end_time,&time_to_wait) == 0)
	return ret_val;
    }
  return 0; 
} 


int 
ACE_VRT_Wrapper::select (int width,
                             fd_set *rfds,
                             fd_set *wfds,
                             fd_set *efds,
                             const ACE_Time_Value *tv)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if (tv == NULL || !vrt->in_test_group())
      return ACE_OS::select_i (width, rfds, wfds, efds, tv);

  int ret_val;
  ACE_Time_Value time_to_wait (vrt->scaled_time_value(*tv));
  ACE_Time_Value end_time = *tv + vrt->current_virtual_time();

  while (1)
    {
      //unlike in the sleep case, here can time-out, or an event can happen 
      ret_val = ACE_OS::select_i (width, rfds, wfds, efds, &time_to_wait);
      if (ret_val != 0 || vrt->wait_longer(end_time, &time_to_wait) == 0)
        return ret_val;
    }
  
  return ret_val;
}

int
ACE_VRT_Wrapper::poll (struct pollfd *pollfds, 
                           u_long len, 
                           const ACE_Time_Value *tv)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if (tv == NULL || !vrt->in_test_group())
    return ACE_OS::poll_i (pollfds, len, tv);

  int ret_val;
  ACE_Time_Value time_to_wait (vrt->scaled_time_value(*tv));
  ACE_Time_Value end_time = *tv + vrt->current_virtual_time();

  while (1)
    {
      //like the select we can get event or time out
      ret_val = ACE_OS::poll_i (pollfds, len, &time_to_wait);
      if (ret_val != 0 || vrt->wait_longer(end_time, &time_to_wait) == 0)
        return ret_val;
    }
  
  return ret_val;
}
      
u_int              
ACE_VRT_Wrapper::alarm (u_int secs)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if ( !vrt->in_test_group())
    return ACE_OS::alarm_i (secs);
  
  //Alarms are controlled by kernel timers therefore cannot be used by a
  // VRT thread
  ACE_NOTSUP_RETURN (0);
}

u_int
ACE_VRT_Wrapper::ualarm (u_int usecs, u_int interval)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if ( !vrt->in_test_group())
    return ACE_OS::ualarm_i (usecs,interval);

  ACE_NOTSUP_RETURN (0);
}

u_int
ACE_VRT_Wrapper::ualarm (const ACE_Time_Value &tv,
                         const ACE_Time_Value &tv_interval)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if ( !vrt->in_test_group())
    return ACE_OS::ualarm_i (tv,tv_interval);

  ACE_NOTSUP_RETURN (0);
}

// if times-out, returns EAGAIN error, Rich's posix bible pp 1873
// if timeout is NULL, the behaviour is unspecified, according to the man page,
// and it can seg-fault here.

int
ACE_VRT_Wrapper::sigtimedwait (const sigset_t *set,
                                   siginfo_t *info,
                                   const ACE_Time_Value *timeout)
{
  ACE_Virtual_Real_Time *vrt = ACE_Virtual_Real_Time::instance();
  if (!vrt->in_test_group())
    return ACE_OS::sigtimedwait_i (set, info, timeout);

  ACE_Time_Value time_to_wait (vrt->scaled_time_value (*timeout));
  ACE_Time_Value end_time = vrt->current_virtual_time() + *timeout;
  
  int ret_val;
  while (1)
    {
      errno = 0;
      ret_val = ACE_OS::sigtimedwait_i (set, info, &time_to_wait);
      
      if (ret_val != -1 ||
          errno != EAGAIN ||
          vrt->wait_longer(end_time, &time_to_wait) == 0)
        break;
    }
  return ret_val; 
}

#endif /* ACE_HAS_VIRTUAL_TIME */
